@@ -1,12 +1,14 @@
-from typing import Tuple, List, Optional, Dict
+from argparse import ArgumentParser, Namespace
+from typing import Tuple, List, Optional
 
 from numpy.random import RandomState
 
 from rlai.actions import Action
 from rlai.agents import Agent
-from rlai.agents.mdp import Human
+from rlai.agents.mdp import Human, StochasticMdpAgent, MdpAgent
 from rlai.environments import Environment
 from rlai.environments.mdp import MdpEnvironment
+from rlai.meta import rl_text
 from rlai.rewards import Reward
 from rlai.states import State
 from rlai.states.mdp import MdpState
@@ -17,99 +19,35 @@ class MancalaState(MdpState):
     State of the mancala game. In charge of representing the entirety of the game state and advancing to the next state.
     """
 
-    @staticmethod
-    def player_1_is_next(
-            picked_pocket,
-            go_again: bool
-    ) -> bool:
-        """
-        Gets whether or not it is player 1's turn next, based on the pocket that was just picked and whether sowing from
-        that pocket has earned another turn.
-
-        :param picked_pocket: Pocket that was picked.
-        :param go_again: Whether or not sowing from the picked pocket earned another turn.
-        :return: True if player 1 is next and False otherwise.
-        """
-
-        return (picked_pocket.player_1 and go_again) or (not picked_pocket.player_1 and not go_again)
-
-    def advance(
-            self,
-            environment: Environment,
-            t: int,
-            a: Action
-    ) -> Tuple[MdpState, int, Reward]:
-        """
-        Advance from the current state given an action.
-
-        :param environment: Environment.
-        :param t: Current time step.
-        :param a: Action.
-        :return: 3-tuple of next state, next time step, and next reward.
-        """
-
-        environment: Mancala
-
-        # pick and sow from pocket
-        picked_pocket = environment.board[a.i]
-        go_again = environment.sow_and_capture(picked_pocket)
-        next_state = MancalaState(
-            mancala=environment,
-            player_1=MancalaState.player_1_is_next(picked_pocket, go_again)
-        )
-
-        t += 1
-
-        # check for termination
-        next_reward = environment.r_none
-        if next_state.terminal:
-            next_reward = environment.get_terminal_reward()
-
-        # if the agent (player 1) does not get to go again, let the environmental agent take its turn(s)
-        elif not go_again:
-            while True:
-
-                environment.player_2.sense(next_state, t)
-
-                # if the environmental agent is human, then render the board for them to see.
-                if isinstance(environment.player_2, Human):
-                    environment.render()
-
-                p2_a = environment.player_2.act(t)
-                picked_pocket = environment.board[p2_a.i]
-                go_again = environment.sow_and_capture(picked_pocket)
-                next_state = MancalaState(
-                    mancala=environment,
-                    player_1=MancalaState.player_1_is_next(picked_pocket, go_again)
-                )
-
-                t += 1
-
-                # check for termination
-                if next_state.terminal:
-                    next_reward = environment.get_terminal_reward()
-                    break
-                # take another turn if earned
-                elif not go_again:
-                    break
-
-        return next_state, t, next_reward
-
     def __init__(
             self,
             mancala,
-            player_1: bool
+            agent_to_sense_state: MdpAgent
     ):
         """
         Initialize the state.
 
         :param mancala: Mancala environment
-        :param player_1: First player (agent).
+        :param agent_to_sense_state: Agent that will sense the state.
         """
 
+        mancala: Mancala
+
+        agent_is_player_1 = agent_to_sense_state != mancala.player_2
+
+        # get the pits ordered from the perspective of the player that will sense the resulting statea
+        if agent_is_player_1:
+            state_pits = mancala.player_1_pockets + [mancala.player_1_store] + mancala.player_2_pockets + [mancala.player_2_store]
+        else:
+            state_pits = mancala.player_2_pockets + [mancala.player_2_store] + mancala.player_1_pockets + [mancala.player_1_store]
+
+        # get state index from the agent that will sense the state
+        state_i_str = '|'.join(str(pit.count) for pit in state_pits)
+        state_i = agent_to_sense_state.get_state_i(state_i_str)
+
         super().__init__(
-            i=mancala.get_state_i(),
-            AA=mancala.get_feasible_actions(player_1),
+            i=state_i,
+            AA=mancala.get_feasible_actions(agent_is_player_1),
             terminal=mancala.is_terminal()
         )
 
@@ -179,44 +117,164 @@ class Pit:
         return f'{self.i}:  Player {1 if self.player_1 else 2}, {self.count}{"*" if self.store else ""}'
 
 
+@rl_text(chapter='Environments', page=1)
 class Mancala(MdpEnvironment):
     """
-    Environment for the mancala game.
+    Environment for the mancala game. This is a simple game with many rule variations, and it provides a greater
+    challenge in terms of implementation and state-space size than the gridworld. I have implemented a fairly common
+    variation summarized below.
+
+    * One row of 6 pockets per player, each starting with 4 seeds.
+    * Landing in the store earns another turn.
+    * Landing in own empty pocket steals.
+    * Game terminates when a player's pockets are clear.
+    * Winner determined by store count.
+
+    A couple hours of Monte Carlo optimization explores more than 1 million states when playing against an equiprobable
+    random opponent.
     """
 
+    @staticmethod
+    def human_player_mutator(
+            environment,
+            **kwargs
+    ):
+        """
+        Change the Mancala environment to let a human play the trained agent.
+
+        :param environment: Environment.
+        :param kwargs: Unused args.
+        """
+
+        environment: Mancala
+        environment.player_2 = Human()
+
+    @classmethod
+    def parse_arguments(
+            cls,
+            args
+    ) -> Tuple[Namespace, List[str]]:
+        """
+        Parse arguments.
+
+        :param args: Arguments.
+        :return: 2-tuple of parsed and unparsed arguments.
+        """
+
+        parsed_args, unparsed_args = super().parse_arguments(args)
+
+        parser = ArgumentParser(allow_abbrev=False)
+
+        parser.add_argument(
+            '--initial-count',
+            type=int,
+            help='Initial seed count in each pit.'
+        )
+
+        parsed_args, unparsed_args = parser.parse_known_args(unparsed_args, parsed_args)
+
+        return parsed_args, unparsed_args
+
     @classmethod
     def init_from_arguments(
             cls,
             args: List[str],
             random_state: RandomState
     ) -> Tuple[Environment, List[str]]:
-        pass
-
-    def get_state_i(
-            self
-    ) -> int:
         """
-        Get the integer identifier for the current board configuration (state). The returned value is guaranteed to be
-        the same for the same board configuration, both throughout the life of the current object as well as after the
-        current object has been pickled for later use (e.g., in checkpoint-based resumption).
+        Initialize an environment from arguments.
 
-        :return: Integer identifier.
+        :param args: Arguments.
+        :param random_state: Random state.
+        :return: 2-tuple of an environment and a list of unparsed arguments.
         """
 
-        state_id_str = '-'.join(str(pit.count) for pit in self.board)
-        if state_id_str not in self.state_id_str_int:
-            self.state_id_str_int[state_id_str] = len(self.state_id_str_int)
+        parsed_args, unparsed_args = cls.parse_arguments(args)
 
-        return self.state_id_str_int[state_id_str]
+        mancala = Mancala(
+            random_state=random_state,
+            player_2=StochasticMdpAgent(
+                'environmental agent',
+                random_state,
+                None,
+                1
+            ),
+            **vars(parsed_args)
+        )
+
+        return mancala, unparsed_args
+
+    def advance(
+            self,
+            state: MdpState,
+            t: int,
+            a: Action,
+            agent: MdpAgent
+    ) -> Tuple[MdpState, Reward]:
+        """
+        Advance from the current state given an action.
+
+        :param state: State to advance.
+        :param t: Current time step.
+        :param a: Action.
+        :param agent: Agent used to generate on-the-fly state identifiers.
+        :return: 2-tuple of next state and next reward.
+        """
+
+        # pick and sow from pocket
+        picked_pocket = self.board[a.i]
+        go_again = self.sow_and_capture(picked_pocket)
+        next_state = MancalaState(
+            mancala=self,
+            agent_to_sense_state=agent if go_again else self.player_2
+        )
+
+        # check for termination
+        next_reward = self.r_none
+        if next_state.terminal:
+            next_reward = self.get_terminal_reward()
+
+        # if the agent (player 1) does not get to go again, let the environmental agent take its turn(s)
+        elif not go_again:
+
+            while True:
+
+                self.player_2.sense(next_state, t)
+
+                # if the environmental agent is human, then render the board for them to see.
+                if isinstance(self.player_2, Human):
+                    self.render()
+
+                p2_a = self.player_2.act(t)
+                picked_pocket = self.board[p2_a.i]
+                go_again = self.sow_and_capture(picked_pocket)
+                next_state = MancalaState(
+                    mancala=self,
+                    agent_to_sense_state=self.player_2 if go_again else agent
+                )
+
+                # check for termination
+                if next_state.terminal:
+                    next_reward = self.get_terminal_reward()
+                    break
+
+                # take another turn if earned
+                elif not go_again:
+                    break
+
+        return next_state, next_reward
 
     def reset_for_new_run(
-            self
+            self,
+            agent: MdpAgent
     ) -> State:
         """
         Reset the game to the initial state.
+
+        :param agent: Agent used to generate on-the-fly state identifiers.
         """
 
-        super().reset_for_new_run()
+        super().reset_for_new_run(agent)
 
         for pocket in self.board:
             if pocket.store:
@@ -224,6 +282,11 @@ class Mancala(MdpEnvironment):
             else:
                 pocket.count = self.initial_count
 
+        self.state = MancalaState(
+            mancala=self,
+            agent_to_sense_state=agent
+        )
+
         return self.state
 
     def render(
@@ -347,28 +410,33 @@ class Mancala(MdpEnvironment):
 
     def __init__(
             self,
-            initial_count: int,
             random_state: RandomState,
+            T: Optional[int],
+            initial_count: int,
             player_2: Agent
     ):
         """
         Initialize the game.
 
-        :param initial_count: Initial count for each pit.
         :param random_state: Random state.
+        :param T: Maximum number of steps to run, or None for no limit.
+        :param initial_count: Initial count for each pit.
         :param player_2: Agent for player 2.
         """
 
+        super().__init__(
+            name='mancala',
+            random_state=random_state,
+            T=T
+        )
+
         self.initial_count = initial_count
         self.player_2 = player_2
 
-        self.state_id_str_int: Dict[str, int] = {}
         self.r_win = Reward(0, 1.0)
         self.r_lose = Reward(1, -1.0)
         self.r_none = Reward(2, 0.0)
 
-        RR = [self.r_win, self.r_lose, self.r_none]
-
         self.player_1_pockets = [
             Pit(True, self.initial_count, False)
             for _ in range(6)
@@ -402,15 +470,3 @@ class Mancala(MdpEnvironment):
         for player_1_pocket, opposing_player_2_pocket in zip(self.player_1_pockets, reversed(self.player_2_pockets)):
             player_1_pocket.opposing_pocket = opposing_player_2_pocket
             opposing_player_2_pocket.opposing_pocket = player_1_pocket
-
-        initial_state = MancalaState(
-            mancala=self,
-            player_1=True
-        )
-
-        super().__init__(
-            name='mancala',
-            random_state=random_state,
-            SS=[initial_state],
-            RR=RR
-        )
