@@ -14,7 +14,7 @@ from rlai.utils import import_function, load_class
 
 @rl_text(chapter='Training and Running Agents', page=1)
 def run(
-        args: List[str]
+        args: List[str] = None
 ) -> Tuple[Optional[str], str]:
     """
     Train an agent in an environment.
@@ -43,6 +43,12 @@ def run(
         help='Fully-qualified type name of environment to train agent in.'
     )
 
+    parser.add_argument(
+        '--planning-environment',
+        type=str,
+        help='Fully-qualified type name of planning environment to train agent in.'
+    )
+
     parser.add_argument(
         '--train-function',
         type=str,
@@ -136,12 +142,6 @@ def run(
         help='N-step update value.'
     )
 
-    train_function_arg_parser.add_argument(
-        '--num-planning-improvements-per-direct-improvement',
-        type=int,
-        help='Number of planning improvements to make for each direct improvement.'
-    )
-
     train_function_arg_parser.add_argument(
         '--new-checkpoint-path',
         type=str,
@@ -192,6 +192,15 @@ def run(
             random_state=random_state
         )
 
+    if parsed_args.planning_environment is not None:
+        planning_environment_class = load_class(parsed_args.planning_environment)
+        train_function_args['planning_environment'], unparsed_args = planning_environment_class.init_from_arguments(
+            args=unparsed_args,
+            random_state=random_state
+        )
+    else:
+        train_function_args['planning_environment'] = None
+
     if len(unparsed_args) > 0:
         raise ValueError(f'Unparsed arguments remain:  {unparsed_args}')
 
